#from msilib.schema import Patch
import pathlib
from os import path
import os
import base64
import sys
import toml
import tomli
import json
from jsonschema import validate as valid
from .parser.Question import CodeRunner
from .parser.Question import MultipleChoice
from .parser.TestCase import TestCase
from .parser.Answer import Answer
from .parser.Question import ShortAnswer
from .parser.Question import Description
from .parser.File import File
from .parser.Category import Category
from jinja2 import Environment, FileSystemLoader
from datetime import datetime

__location__ = os.path.realpath(
    os.path.join(os.getcwd(), os.path.dirname(__file__)))

def validate(file, dir):

    with open(os.path.join(__location__, "schema.json")) as f:
        schema = json.loads(f.read())
        tomlFile = tomli.loads(file)
        tomlFile = json.dumps(tomlFile)
        tomlFile = json.loads(tomlFile)
        valid(instance=tomlFile, schema=schema)
        


def getQuestion(dir):
    print("[{}] Loading question from {}".format(datetime.now().strftime("%H:%M:%S"), dir))
    if dir.suffix == ".cr":
        print("[{}]   Question format: CodeRunner".format(datetime.now().strftime("%H:%M:%S")))
        newQuestion = CodeRunner()
    elif dir.suffix == ".mc":
        print("[{}]   Question format: Multiple choice".format(datetime.now().strftime("%H:%M:%S")))
        newQuestion = MultipleChoice()
    elif dir.suffix == ".sa":
        newQuestion = ShortAnswer()
    elif dir.suffix == ".de":
        newQuestion = Description()
    else:
        print(dir.suffix)
    for p in dir.iterdir():
        if not p.is_file():
            for sf in p.iterdir():
                if sf.is_file():
                    with open(sf, "rb") as f:
                        content = f.read()
                        newFile = File(sf.name, '/')
                        newFile.setContent(base64.b64encode(content))
                        if sf.suffix in ".png,.jpg,.jpeg":
                            newQuestion.addImage(newFile)
                        else:
                            newQuestion.addFile(newFile)
            continue
        with open(p) as f:
            if p.suffix == ".py":
                print("[{}]   Input format: .py".format(datetime.now().strftime("%H:%M:%S")))

                answer = f.read()
                newQuestion.setAnswer(answer)

            elif p.suffix == ".md":
                print("[{}]   Input format: .md".format(datetime.now().strftime("%H:%M:%S")))

                prompt = f.read()
                newQuestion.setPrompt(prompt)
            elif p.suffix == ".txt":
                feedback = f.read()
                newQuestion.setFeedback(feedback)

            elif p.suffix == ".toml":
                validate(f.read(), dir)
                f.seek(0)
                fileRead = f.readlines()
                print("[{}]   Input format: .toml".format(datetime.now().strftime("%H:%M:%S")))

                cases = []
                lines = iter(fileRead)
                for line in lines:
                    if line.strip() == "":
                        continue
                    if "[[testcases]]" in line:
                        cases.append(TestCase())
                    elif "[[answers]]" in line:
                        cases.append(Answer())
                    else:
                        splitLine = line.split("=")
                        #take what's before an equal sign and that's our variable name we are manipulating
                        attributeName = splitLine[0].strip()
                        attributeValue = None
                        #get everything inbetween the three ''' '''
                        if "'''" not in line:
                            attributeValue = splitLine[1].strip()
                        else:
                            equator = splitLine[1].split("'''")
                            attributeValue = equator[1]
                            if len(equator) == 2:
                                while True:
                                    nextLine = next(lines)
                                    if "'''" in nextLine:
                                        attributeValue += nextLine.split("'''")[0]
                                        break
                                    attributeValue += nextLine
                        if len(cases) > 0:
                            cases[-1].__setattr__(attributeName, attributeValue)
                        else:
                            if attributeName == "tags":
                                attributeValue = attributeValue.split(",")
                            newQuestion.__setattr__(attributeName, attributeValue)
                newQuestion.setCases(cases)          
    return newQuestion


def IterateChildren(dir, cat, globLocations, globPattern):
    #if it's a question
    if(dir.suffix != ""):
        [questionName, type] = dir.name.split(".")
        question = getQuestion(dir)
        question.setTitle(questionName)
        cat[-1].addQuestion(question)
    else:       
        cat.append(createCategory(dir))
        #checkSubDirectories
        for p in dir.iterdir():
            if p not in globLocations["whitelist"] and globPattern["IterateChildren"] and p not in globLocations["blacklist"]:

                cat = (IterateChildren(p, cat, globLocations, globPattern))
    return cat
    


def createCategory(location, startingPart = 1):
    if len(location.parts) == 1:
       catName = str(pathlib.Path(location.parts[0]))
    else:
        catName = str(pathlib.Path(*location.parts[startingPart:]))
    return Category(catName)

def importQuestions(root, globPattern):
    globLocations = {
        "whitelist": list(root.rglob(globPattern["exportGlob"])),
        "blacklist": list(root.rglob(globPattern["blackListGlob"]))
    } 
    if globPattern["blackListGlob"] == "":
        globLocations["blacklist"] = []
    if len(globLocations["whitelist"]) == 0:
        return ""
    globLocations["whitelist"] = [x for x in globLocations["whitelist"] if (x.is_dir() and str(x.parent.suffix) == "")]
    numQuestions = len([x for x in globLocations["whitelist"] if x.suffix != ""])

    smallestParent = min([len(x.parts) for x in globLocations["whitelist"]])-1

    categories = []
    #if all in the glob are questions then
    if numQuestions == len(globLocations["whitelist"]):
        categories.append(createCategory(globLocations["whitelist"][0].parent, smallestParent-1))


    for pa in globLocations["whitelist"]:
        #if path contains anything from any of the blacklist strings
        for location in globLocations["blacklist"]:
            if str(location) in str(pa):
                break
        else:
            categories = (IterateChildren(pa, categories, globLocations, globPattern))  
            continue
        break

    file_loader = FileSystemLoader(path.join(path.dirname(__file__), 'templates'))

    env = Environment(loader=file_loader)

    #get our questions/answers
    #categories = getCategories(root)
    quizTemplate = env.get_template('quiz.xml')

    for cat in categories:
        cat.convertQuestions(env)

    output = quizTemplate.render(categories=categories)


    return output


import pathlib
from jinja2 import Environment, FileSystemLoader
import re
from datetime import datetime


class Question:
    def __init__(self):
        self.title = "null"
        self.prompt = "null"
        self.files = []
        self.images = []
        self.tags = []
        self.defaultgrade = 1
        self.penalty = 0
        self.answerlines = 15
        self.precheck = 0
        self.answerpreload = ''''''
        self.casesensitivity = 0
        

    def setTitle(self, title):
        self.title = title

    def setPrompt(self, prompt):
        self.prompt = prompt

    def setGrade(self, grade):
        self.grade = grade
    
    def addFile(self, file):
        self.files.append(file)

    def addImage(self, file):
        self.images.append(file)

class MultipleChoice(Question):
    _XML_NAME = "multiplechoice.xml"
    def __init__(self):
        super(MultipleChoice, self).__init__()
        self.shuffle = 'true'
        self.answernumbering = 'abc'
        self.singleAnswer = 'true'

    def setCases(self, answers):
        self.answers = answers

class Description(Question):
    _XML_NAME = "description.xml"
    def __init__(self):
        super(Description, self).__init__()

    def setFeedback(self, feedback):
        self.feedback = feedback

class ShortAnswer(Question):
    _XML_NAME = "shortanswer.xml"
    def __init__(self):
        super(ShortAnswer, self).__init__()
        
    def setCases(self, answers):
        self.answers = answers

    def setFeedback(self, feedback):
        self.feedback = feedback



class CodeRunner(Question):
    _XML_NAME = "coderunner.xml"
    def __init__(self):
        super(CodeRunner, self).__init__()
        self.coderunnertype = "python3_w_input"
        self.template = None

    def setAnswer(self, answer):
        self.answer = answer

    def setCases(self, testcases):
        self.testcases = testcases

    def testTestCases(self):
        #move the input part to a previous line
        #convert Answer inputs to console
        #testCaseCode = self.answer.replace('input(', 'print(')
        #print(testCaseCode)
        for loopNum, test in enumerate(self.testcases):
            test.testExpected(self, loopNum)


from pathlib import Path
from .moodle_compile.compile import importQuestions
from datetime import datetime

import typer
from typer import Argument, Option

app = typer.Typer(
    add_completion=False,
)

def build(root: Path = Path("questions"), globPattern = {
        "exportGlob": "",
        "blackListGlob": "None",
        "IterateChildren": False
    }):

    print("============")
    print("Finale 0.1.8")
    print("============")
    print()

    print("[{}] Starting Finale".format(datetime.now().strftime("%H:%M:%S")))

    #targeted export by glob & exclusion by glob
    #export Children decides whether or not you want to export 
    #children of glob pattern category
    #make sure blackListGlob isn't accidentally blacklisting everything
    #use Null to match with nothing and "" to match with everything
    xml = importQuestions(root, globPattern)
    with open("out.xml", "w") as f:
        f.write(xml)

@app.command(
    "new",
    help="Generates a blank question template in the current working directory",
)
def new(
        qpath: str = Argument(
            ...,
            help="The question type to generate",
        ),
        name: str = Argument(
            ...,
            help="The name of the question to generate",
        ),
        qtype: str = Argument(
            ...,
            help="The question type to generate",
        ),
):
    createBlankQuestion(qpath, name, qtype)

def createBlankQuestion(qpath, name, qtype):
    qpath = f'{qpath}/{name}.{qtype}'
    qloc = Path('questions').joinpath(qpath)

    #create support files folder
    createPath = qloc.joinpath('Support Files')
    createPath.mkdir(parents=True, exist_ok=True)
    
    #create a prompt.md for all
    qloc.joinpath('prompt.md').touch(exist_ok=True)
    
    #create a Feedback.txt for .sa and .de
    if qtype in ["sa", "de"]:
        qloc.joinpath('Feedback.txt').touch(exist_ok=True)

    #create testcases.toml and answers.py for .cr
    if qtype == "cr":
        qloc.joinpath('testcases.toml').touch(exist_ok=True)
        qloc.joinpath('answers.py').touch(exist_ok=True)

    #create answers.toml for .sa and .mc
    if qtype in ["sa", "mc"]:
        qloc.joinpath('answers.toml').touch(exist_ok=True)
    
    
if __name__ == '__main__':
    app()



class File:
    def __init__(self, name, path, encoding, contents):
        self.name = name
        self.path = path
        self.encoding = encoding
        self.content = content

    def __init__(self, name, path):
        self.name = name
        self.path = path
        self.encoding = 'base64'
        self.content = None

    def setContent(self, newContent):
        self.content = newContent.decode("utf-8") 



class Answer:
    def __init__(self, fraction, text, feedback):
        self.fraction = fraction
        self.text = text
        self.feedback = feedback

    def __init__(self):
        self.fraction = None
        self.text = ""
        self.feedback = ""

    def setfraction(self, fraction):
        self.fraction = fraction

    def settext(self, text):
        self.text = text

    def setfeedback(self, feedback):
        self.feedback = feedback



from .Question import CodeRunner

class Category:
    def __init__(self, name, questions):
        self.name = name
        self.questions = questions

    def __init__(self, name):
        self.name = name
        self.questions = []

    def addQuestion(self, newQuestion):
        self.questions.append(newQuestion)

    def convertQuestions(self, env):
        questionTemplate = []
        #This is where we can test each questions testcases 
        for questionOBJ in self.questions:
            if(isinstance(questionOBJ, CodeRunner)):
                questionOBJ.testTestCases()
            template = env.get_template(questionOBJ._XML_NAME)

            questionTemplate.append(template.render(question=questionOBJ))

        self.questions = questionTemplate




import sys
from io import StringIO
from contextlib import redirect_stdout

class TestCase:
    def __init__(self, example, display, testcode, stdin, expected):
        self.example = example
        self.display = display
        self.testcode = testcode
        self.stdin = stdin
        self.expected = expected

    def __init__(self):
        self.example = None
        self.display = None
        self.testcode = ""
        self.stdin = ""
        self.expected = ""

    def testExpected(self, question, testNumber):
        stds = self.stdin.split("\n")
        splitAnswer = question.answer.split("\n")

        #add a print statement for the input, since setting the sys.stdin doesn't print the user input
        for i in range(len(splitAnswer)):
            if "input(" in splitAnswer[i]:
                splitAnswer[i] += '\nprint("'+ stds.pop(0) + '")'
        answer = "\n".join(splitAnswer)

        #runs the answer using the IO and then saves it in a temporary console
        sys.stdin = StringIO(self.stdin)
        with redirect_stdout(StringIO()) as tempConsole:
            exec(answer)

        knownExpected = tempConsole.getvalue()[:-1]
        if self.expected == "":
            #add expected output using answer
            self.expected = knownExpected
        else:
            #run expected using answer and compare
            if(knownExpected != self.expected):
                raise Exception(f"Testcase doesn't have correct expected output for CodeRunner question titles='{question.title}'. Testcase #{testNumber+1}")
